/-

Type Soundness in Call-by-value, Simply Typed Lambda Calculus with References

In this homework, you'll prove soundness for a version of the call-by-value,
simply-typed lambda calculus which has been extended to include references.

References provide *mutable* updates to variables. They can be allocated with
"ref", dereferenced with `!`, and assigned to with `<-`. At runtime, references
are represented as "addresses" into a list of terms. As usual, we will prove
that well-typed programs don't get stuck. For references, this requires that
well-typed terms don't contain pointers to unallocated addresses. We will also
require that the type of the term stored at a given address never changes.

In order to deal with references to the heap, we need to add addresses
(represented in the semantics as `term.addr n`). Programmers writing programs
in our language won't write these directly, but when a reference is allocated
with `ref <term>`, the term is stored in the heap and the reference steps to
the address of that term.

We encourage you to start early, and to consult the lecture notes--many of the
proofs will be similar to proofs from lecture.

Total points: 120

-/

import struct_tact

namespace stlc_ref

/-- Syntax -/
inductive term : Type
| bool   : bool → term
| int    : int → term
| var    : string → term
| app    : term → term → term
| lam    : string → term → term
| addr   : nat → term
| ref    : term → term
| deref  : term → term
| assign : term → term → term

instance bool_to_term : has_coe bool term :=
⟨ λ b, term.bool b ⟩

instance int_to_term : has_coe int term :=
⟨ λ i, term.int i ⟩

instance var_to_term : has_coe string term :=
⟨ λ s, term.var s ⟩

infixr @@ := term.app
prefix ! := term.deref
notation `\` x `,` y := (term.lam x y)
def ref := term.ref
infix <= := term.assign

/- Substitution -/

/-- e1[e2/x] = e3 -/
inductive is_subst : term → term → string →
                     term → Prop
| bool:
    forall b e x,
      is_subst (term.bool b) e x
               (term.bool b)
| int:
    forall i e x,
      is_subst (term.int i) e x
            (term.int i)
| var_same:
    forall e x,
      is_subst (term.var x) e x
            e
| var_diff:
    forall e x1 x2,
      x1 ≠ x2 →
      is_subst (term.var x1) e x2
               (term.var x1)
| app:
    forall e1 e2 e x e1' e2',
      is_subst e1 e x e1' →
      is_subst e2 e x e2' →
      is_subst (term.app e1 e2) e x
               (term.app e1' e2')
| lam_same:
    forall e1 x e,
      is_subst (term.lam x e1) e x
               (term.lam x e1)
| lam_diff:
    forall e1 x1 x2 e e1',
      x1 ≠ x2 →
      is_subst e1 e x2 e1' →
      is_subst (term.lam x1 e1) e x2
               (term.lam x1 e1')
| addr :
    forall a e x,
      is_subst (term.addr a) e x
               (term.addr a)
| ref :
    forall r e x r',
      is_subst r e x r' →
      is_subst (term.ref r) e x (term.ref r')
| deref :
    forall r e x r',
      is_subst r e x r' →
      is_subst (term.deref r) e x (term.deref r')
| assign:
    forall e1 e2 e x e1' e2',
      is_subst e1 e x e1' →
      is_subst e2 e x e2' →
      is_subst (term.assign e1 e2) e x
               (term.assign e1' e2')

/--

[PROBLEM 1 (5 points)]:

  Prove that our is_substitution relation is deterministic.

-/
lemma is_subst_det:
  forall e1 e2 x e3,
    is_subst e1 e2 x e3 →
    forall e3',
      is_subst e1 e2 x e3' →
      e3 = e3' :=
/- About 44 tactics -/
begin
  admit
end
/- END PROBLEM 1 -/

/-- What does it mean for a variable to be is_free in an term?
    There shouldn't be any surprises here, since references don't
    affect whether a variable is bound or not. -/
inductive is_free : term → string → Prop
| var:
    forall x,
      is_free (term.var x) x
| app_l:
    forall x e1 e2,
      is_free e1 x →
      is_free (term.app e1 e2) x
| app_r:
    forall x e1 e2,
      is_free e2 x →
      is_free (term.app e1 e2) x
| lam:
    forall x1 x2 e,
      is_free e x1 →
      x1 ≠ x2 →
      is_free (term.lam x2 e) x1
| ref :
    forall x r,
      is_free r x →
      is_free (term.ref r) x
| deref :
    forall x r,
      is_free r x →
      is_free (term.deref r) x
| assign_l:
    forall x e1 e2,
      is_free e1 x →
      is_free (term.assign e1 e2) x
| assign_r:
    forall x e1 e2,
      is_free e2 x →
      is_free (term.assign e1 e2) x.

/- If a variable isn't `is_free` in an term,
   substituting that variable doesn't change the term. -/
lemma is_subst_only_is_free:
  forall e1 e2 x e3,
    is_subst e1 e2 x e3 →
    ¬ is_free e1 x →
    e1 = e3 :=
begin
  intros e1 e2 x e3 e1_subst,
  induction e1_subst; intros; try { trivial },
  { exfalso, apply a,
    constructor,
  },
  { congr,
    apply ih_1,
    delta not, intros,
    apply a_2,
    apply is_free.app_l,
    assumption,
    apply ih_2,
    delta not, intros,
    apply a_2,
    apply is_free.app_r,
    assumption,
  },
  { congr,
    apply ih_1,
    delta not, intros,
    apply a_2,
    apply is_free.lam; try { assumption },
    apply ne.symm; assumption,
   },
  { congr, apply ih_1,
    delta not; intros,
    apply a_1,
    constructor; assumption,
  },
  { congr, apply ih_1,
    delta not; intros,
    apply a_1,
    constructor; assumption,
  },
  { congr,
    apply ih_1,
    delta not, intros,
    apply a_2,
    apply is_free.assign_l,
    assumption,
    apply ih_2,
    delta not, intros,
    apply a_2,
    apply is_free.assign_r,
    assumption,
  },
end


/-- Closed terms have no is_free variables -/
def is_closed (e: term) : Prop :=
  forall x, ¬ is_free e x

/- These are a bunch of boring facts about closed terms.
   We've completed the proofs, but look over them because
   they are will be useful later. -/
lemma closed_app_intro:
  forall e1 e2,
    is_closed e1 →
    is_closed e2 →
    is_closed (e1 @@ e2) :=
begin
  unfold is_closed; delta not; intros,
  cases a_2,
  apply a; assumption,
  apply a_1; assumption,
end

lemma closed_app_inv:
  forall e1 e2,
    is_closed (e1 @@ e2) →
    is_closed e1 /\ is_closed e2 :=
begin
  unfold is_closed; delta not; intros,
  split; intros,
  { apply a,
    apply is_free.app_l; assumption },
  { apply a,
    apply is_free.app_r; assumption }
end

lemma closed_lam_intro:
  forall x e,
    (forall y, y ≠ x → ¬ is_free e y) →
    is_closed (\x, e) :=
begin
  unfold is_closed; delta not; intros,
  cases a_1, apply a; assumption,
end

lemma closed_lam_inv:
  forall x e,
    is_closed (\x, e) →
    (forall y, y ≠ x → ¬ is_free e y) :=
begin
  unfold is_closed; delta not; intros,
  have y_free : is_free (\x, e) y,
  constructor; assumption,
  apply a; assumption,
end

lemma closed_ref_intro:
  forall e,
    is_closed e →
    is_closed (ref e) :=
begin
  unfold is_closed; delta not; intros,
  cases a_1, apply a; assumption
end

lemma closed_ref_inv:
  forall e,
    is_closed (ref e) →
    is_closed e :=
begin
  unfold is_closed; delta not; intros,
  apply a, constructor; assumption,
end

lemma closed_deref_intro:
  forall e,
    is_closed e →
    is_closed (! e) :=
begin
  unfold is_closed; delta not; intros,
  cases a_1, apply a; assumption,
end

lemma closed_deref_inv:
  forall e,
    is_closed (! e) →
    is_closed e :=
begin
  unfold is_closed, delta not; intros,
  apply a, constructor; assumption,
end

lemma closed_assign_intro:
  forall e1 e2,
    is_closed e1 →
    is_closed e2 →
    is_closed (e1 <= e2) :=
begin
  unfold is_closed; delta not; intros,
  cases a_2,
  apply a; assumption,
  apply a_1; assumption
end

lemma closed_assign_inv:
  forall e1 e2,
    is_closed (e1 <= e2) →
    is_closed e1 ∧ is_closed e2 :=
begin
  unfold is_closed; delta not; intros,
  split; intros,
  apply a, constructor; assumption,
  apply a, apply is_free.assign_r; assumption,
end

/--

[PROBLEM 2 (5 points)]:

  Prove that closed-ness is preserved by substitution.
  This proof should use many of the lemmas defined above.

-/
lemma is_subst_pres_closed:
  forall e1 e2 x e3,
    is_subst e1 e2 x e3 →
    is_closed e1 →
    is_closed e2 →
    is_closed e3 :=
/- About 24 tactics -/
begin
  admit
end

/- END PROBLEM 2 -/


/-

Here we define "heaps", which our references will index into.

A heap is just a list of termessions (in our uses later below,
they will always be values) which can be indexed into.

-/

@[reducible] def heap := list term

/-- The empty heap is just the empty list. -/
def heap.empty : heap := []

/-
We define lookup in terms of the `nth` function.

nth returns an option (consider why!), but
we will prove that we won't end up in the case where
it returns none in the below code.
-/

@[irreducible] def heap.lookup (h : heap) (addr : nat) : term :=
option.get_or_else (h.nth addr) (term.bool tt)

def heap.length (h : heap) : nat :=
list.length h

/- `snoc` is `cons` backwards. It adds an element to the end of a list. -/
def snoc {α : Type} : list α → α → list α
| [] x := x :: []
| (h :: ts) x := h :: snoc ts x

/- We will need some boring lemmas about [snoc]. We've completed the
proofs for you, but look over them since you'll need them later.  -/

lemma length_snoc :
  forall (a : Type) (l : list a) n,
  list.length (snoc l n) = 1 + (list.length l) :=
begin
  induction_on l; simp [list.length, snoc, *] at *,
end

lemma nth_lt_snoc :
forall (α : Type) (l : list α) x n,
  n < list.length l →
  list.nth l n = list.nth (snoc l x) n :=
begin
  induction_on l; intros,
  { cases a, },
  { cases n; simp [snoc, list.nth],
    apply ih_1,
    simp [has_add.add, nat.add] at a_2,
    apply nat.lt_of_succ_lt_succ; assumption,
  }
end

lemma nth_eq_snoc :
forall (a : Type) (l : list a) x,
  list.nth (snoc l x) (list.length l) = some x :=
begin
  induction_on l; intros,
  { simp [snoc, list.nth] },
  { simp [snoc], rw add_comm,
    rw list.nth.equations._eqn_3,
    apply ih_1,
  }
end

/- To update the heap, we use the `replace` function, which replaces
   the contents of a cell at a particular index. -/
def replace {α : Type} : nat → α → list α → list α
| n _ [] := []
| 0 x (h :: t) := x :: t
| (n' + 1) x (h :: t) := h :: replace n' x t

/-* Of course, we also need some boring lemmas about [replace], which
    are also fairly straightforward to prove. -/

lemma replace_nil : forall A n (x:A),
  replace n x [] = [] :=
begin
  intros; cases n; simp [replace]
end

lemma length_replace : forall A (l:list A) n x,
  list.length (replace n x l) = list.length l :=
begin
  induction_on l; intros;
  cases n; simp [list.length, replace, *] at *,
end

lemma nth_replace_eq :
forall (h : heap) n t,
  n < list.length h →
  list.nth (replace n t h) n = some t :=
begin
  induction_on h; intros,
  { simp [list.length] at *, cases a, },
  { cases n; simp [replace, list.nth],
    apply ih_1,
    simp [list.length, *] at a_2,
    rw add_comm at a_2,
    simp [has_add.add, nat.add] at a_2,
    apply nat.lt_of_succ_lt_succ; assumption,
  }
end

lemma lookup_replace_eq : forall n h t,
  n < list.length h →
  heap.lookup (replace n t h) n = t :=
begin
  unfold heap.lookup; intros,
  rw nth_replace_eq; try { assumption },
  simp [option.get_or_else],
end

lemma nat_succ_neq_neq :
  forall n m,
    nat.succ n ≠ nat.succ m →
    n ≠ m :=
begin
  delta ne not; intros,
  apply a,
  simp * at *,
end

/--

[PROBLEM 3 (5 points)]:

  Prove that replacing a value in a heap doesn't
  affect lookups for different values.

  We woud like to directly prove `lookup_replace_neq`,
  but the definition of `heap.lookup` using `nth` and `replace`
  means the theorem we want to prove is not directly inductive.

  We can instead write down another lemma `nth_replace_neq`
  which is inductive and then use it to easily prove the lemma
  we are interested in.

  Hint: You should use nat_succ_neq_neq.
 -/
lemma nth_replace_neq :
  forall (α : Type) (t : α) (h : list α) (addr1 addr2 : nat),
  addr1 ≠ addr2 →
  list.nth (replace addr1 t h) addr2 = list.nth h addr2 :=
/- About 12 tactics -/
begin
  admit
end

lemma lookup_replace_neq : forall h t a1 a2,
  a1 ≠ a2 →
  heap.lookup (replace a1 t h) a2 = heap.lookup h a2 :=
/- About 4 tactics -/
begin
  admit
end

/- END PROBLEM 3 -/

lemma heap_lookup_snoc :
  forall (l : heap) x,
    heap.lookup (snoc l x) (list.length l) = x :=
begin
  intros,
  unfold heap.lookup,
  rw nth_eq_snoc,
  simp [option.get_or_else],
end

lemma heap_lt_lookup_snoc :
  forall (l : heap) x n,
    n < list.length l →
    heap.lookup (snoc l x) n = heap.lookup l n  :=
begin
  intros,
  unfold heap.lookup,
  rw <- nth_lt_snoc; assumption,
end

/--
Now that we have heaps, let's define our semantics!

Since we're writing a call-by-value semantics,
we first need to define "values".

-/
inductive is_value : term → Prop
| lam  : forall x e, is_value (\ x, e)
| int  : forall i : int, is_value i
| bool : forall b : bool, is_value b
| addr : forall n, is_value (term.addr n)

/-

Our step relation includes heaps as well as termessions, since heaps
can change. Look *carefully* over this step relation and make sure you
understand every rule!  Really, you'll need to grok this to finish the
homework.

-/
inductive step_cbv : heap → term → heap → term → Prop
| app_left :
    forall h h' e1 e1' e2,
      step_cbv h e1
               h' e1' →
      step_cbv h (e1 @@ e2)
               h' (e1' @@ e2)
| app_right :
    forall h h' e1 e2 e2',
      is_value e1 →
      step_cbv h e2
               h' e2' →
      step_cbv h (e1 @@ e2)
               h' (e1 @@ e2')
| app :
    forall h x e1 e2 e1',
      is_value e2 →
      is_subst e1 e2 x e1' →
      step_cbv h ((\ x, e1) @@ e2)
               h e1'
| ref :
    forall h h' e e',
      step_cbv h e
               h' e' →
      step_cbv h (ref e)
               h' (ref e')
| ref_value :
    forall h e,
      is_value e →
      step_cbv h (ref e)
               (snoc h e) (term.addr (list.length h))
| deref :
    forall h h' e e',
      step_cbv h e
               h' e' →
      step_cbv h (! e)
               h' (! e')
| deref_addr :
    forall (h : heap) a,
      a < list.length h →
      step_cbv h (! (term.addr a))
               h (h.lookup a)
| assign_left :
    forall h h' e1 e1' e2,
      step_cbv h e1
               h' e1' →
      step_cbv h (e1 <= e2)
               h' (e1' <= e2)
| assign_right :
    forall h h' e1 e2 e2',
      is_value e1 →
      step_cbv h e2
               h' e2' →
      step_cbv h (e1 <= e2)
               h' (e1 <= e2')
| assign :
    forall h a e,
      is_value e →
      a < list.length h →
      step_cbv h (term.addr a <= e)
               (replace a e h) (term.bool tt)

notation h1 `;` e1 `==>` h2 `;` e2 :=
  (step_cbv h1 e1 h2 e2)

/- Any number of steps -/
inductive star_cbv : heap → term → heap → term → Prop
| refl:
    forall h e,
      star_cbv h e h e
| step:
    forall h1 e1 h2 e2 h3 e3,
      step_cbv h1 e1 h2 e2 →
      star_cbv h2 e2 h3 e3 →
      star_cbv h1 e1 h3 e3

notation h1 `;` e1 `==>*` h2 `;` e2 :=
  (star_cbv h1 e1 h2 e2)

/- Let's talk about types! -/

/- We'll need to add a type for references to the set of types we saw lecture. -/

inductive typ : Type
| bool : typ
| int  : typ
| fn   : typ → typ → typ
| ref  : typ → typ.

infixr `~>`:50 := typ.fn

/- An environment maps variables to types. Make sure you understand
the difference between this and a heap, which maps indices to
terms! -/

@[reducible] def env : Type := string → option typ

/- E0 is the empty environment -/
@[irreducible] def env.empty : env :=
λ _, none.

def env.lookup (e : env) (x : string) : option typ :=
e x

/-

Update an environment with a new variable and type.

NOTE: Environments are different from heaps!
      We change heaps with snoc and replace.
      We change environments with extend.

-/
@[irreducible]
def extend (e: env) (x : string) (t : typ) : env :=
  λ y,
    if x = y then
      some t
    else
      e y.

lemma lookup_extend_neq :
  forall e x x' t,
  x ≠ x' →
  env.lookup (extend e x t) x' = env.lookup e x' :=
begin
  intros,
  unfold env.lookup extend,
  break_if,
end

/- In addition to our usual environments,
   we also need to type our heaps in order
   to type references. -/

/- A heap type is just a list of types. -/
def heap_types := list typ.

/-- A empty set of types, corresponds to the types for an empty heap. -/
@[irreducible] def heap_types.empty : heap_types := []

/- lookup_typ works just like lookup. -/
@[irreducible] def heap_types.lookup (h : heap_types) (addr : nat) :=
option.get_or_else (list.nth h addr) typ.bool

/- I will be useful later on. -/
lemma heap_types_lookup_snoc :
  forall (l : heap_types) x,
    heap_types.lookup (snoc l x) (list.length l) = x :=
begin
  intros,
  unfold heap_types.lookup,
  rw nth_eq_snoc,
  simp [option.get_or_else],
end

lemma heap_types_lt_lookup_snoc :
  forall (l : heap_types) x n,
    n < list.length l →
    heap_types.lookup (snoc l x) n = heap_types.lookup l n  :=
begin
  intros,
  unfold heap_types.lookup,
  rw <- nth_lt_snoc; assumption
end

@[irreducible] def heap_types.length (h : heap_types) : nat :=
list.length h

/- What does it mean for a term to be well-typed?
   The first 5 constructors are the same as those
   in the STLC without references.
-/
@[auto] inductive typed : env → heap_types → term → typ → Prop
| bool :
    forall env ht b,
      typed env ht (term.bool b) typ.bool
| int :
    forall env ht i,
      typed env ht (term.int i) typ.int
| var :
    forall (env : env) ht x t,
      env.lookup x = some t →
      typed env ht (term.var x) t
| app :
    forall env ht e1 e2 tA tB,
      typed env ht e1 (tA ~> tB) →
      typed env ht e2 tA →
      typed env ht (e1 @@ e2) tB
| lam :
    forall env ht x e tA tB,
      typed (extend env x tA) ht e tB →
      typed env ht (\x, e) (tA ~> tB)
| addr :
    forall env (ht : heap_types) a,
      a < list.length ht →
      typed env ht (term.addr a) (typ.ref (ht.lookup a))
| ref :
    forall env ht e t,
      typed env ht e t →
      typed env ht (ref e) (typ.ref t)
| deref :
    forall env ht e t,
      typed env ht e (typ.ref t) →
      typed env ht (! e) t
| assign :
    forall env ht e1 e2 t,
      typed env ht e1 (typ.ref t) →
      typed env ht e2 t →
      typed env ht (e1 <= e2) typ.bool.

/- Q: What does it mean for a heap to be well-typed?

   A: The heap must be the same length as the heap type, and the term
      stored at any valid address in the heap (i.e. any address less than
      the length of the heap) should have the type it has in the heap type.
-/

def heap_typed (ht : heap_types) (h : heap) :=
  list.length h = list.length ht ∧
  forall (a : nat), a < list.length h →
    typed env.empty ht (h.lookup a) (ht.lookup a).


/- We want to prove soundness. As usual, we'll prove *progress* and
*preservation*. We'll start with progress, since it's easier. -/


/-- In our calculus, a number of types have canonical forms: if a
member of that type is a value, it must have a particular form. These
will be useful in proving progress. -/

lemma canon_bool:
  forall env ht e,
    is_value e →
    typed env ht e typ.bool →
    exists b, e = term.bool b :=
begin
  intros,
  cases a; cases a_1; constructor; trivial,
end

lemma canon_int:
  forall env ht e,
    is_value e →
    typed env ht e typ.int →
    exists i, e = term.int i :=
begin
  intros, cases a; cases a_1; constructor, trivial,
end

lemma canon_fun:
  forall env ht e tA tB,
    is_value e →
    typed env ht e (tA ~> tB) →
    exists x, exists b, e = \x, b :=
begin
  intros,
  cases a; cases a_1,
  existsi _, existsi _,
  reflexivity,
end

lemma canon_ref:
  forall env ht e t,
    is_value e →
    typed env ht e (typ.ref t) →
    exists a, e = term.addr a :=
begin
  intros,
  cases a; cases a_1; existsi _; trivial,
end

/- We can always substitute. You'll destruct the value of can_is_subst on
particular arguments somewhere in the progress proof. -/
lemma can_subst:
  forall e1 e2 x,
    exists e3, is_subst e1 e2 x e3 :=
begin
  induction_on e1; intros,
  { repeat { constructor } },
  { repeat { constructor } },
  { by_cases (a = x),
    subst h,
    constructor,
    apply is_subst.var_same,
    constructor,
    apply is_subst.var_diff,
    simp * at *,
  },
  { specialize (ih_1 _ _),
    specialize (ih_2 _ _),
    cases ih_1,
    cases ih_2,
    constructor, constructor,
    assumption,
    assumption,
  },
  { specialize (ih_1 _ _),
    cases ih_1,
    by_cases (a = x),
    subst h,
    constructor,
    apply is_subst.lam_same,
    constructor,
    apply is_subst.lam_diff,
    simp * at *,
    assumption,
  },
  { constructor,
    constructor, },
  { specialize (ih_1 _ _),
    cases ih_1,
    constructor,
    constructor; assumption,
  },
  { specialize (ih_1 _ _),
    cases ih_1,
    constructor,
    constructor,
    assumption,
  },
  { specialize (ih_1 _ _ ),
    specialize (ih_2 _ _),
    cases ih_1,
    cases ih_2,
    constructor,
    constructor; assumption,
  },
end

/-*

[PROBLEM 4 (30 points)]:

Prove progress for our calculus: given a well-typed heap, a well-typed
sterm can either step or is a value.  We've started the proof for you.

-/
lemma progress:
  forall ht h e t
    (e_typed : typed env.empty ht e t),
    heap_typed ht h →
    (exists e' h', h; e ==> h'; e') \/ is_value e :=
/- About 90 tactics -/
begin
  induction_on e_typed; subst eq,
  { admit },
  { admit },
  { admit },
  { admit },
  { admit },
  { admit },
  { admit },
  { admit },
  { admit },
end
/- END PROBLEM 4 -/

/-
Next, we'll prove preservation. This is going to be a little harder;
we're going to need to develop some machinery around environments and
heap types. -/

/- Equivalence of environments -/
def env_equiv (e1 e2 : env) : Prop :=
  forall s, e1.lookup s = e2.lookup s

/- Some lemmas about this equivalence relation. -/

/- Reflexivity -/
lemma env_equiv_refl:
  forall env,
    env_equiv env env :=
begin
  unfold env_equiv; intros, constructor,
end

/- Symmetry -/
lemma env_equiv_sym:
  forall e1 e2,
    env_equiv e1 e2 →
    env_equiv e2 e1 :=
begin
  unfold env_equiv; intros; simp * at *
end

/-- Transitivity -/
lemma env_equiv_trans:
  forall e1 e2 e3,
    env_equiv e1 e2 →
    env_equiv e2 e3 →
    env_equiv e1 e3 :=
begin
  unfold env_equiv; intros,
  simp * at *,
end

/-- Extending environments with the same value preserves equivalence. -/
lemma env_equiv_extend:
  forall env1 env2 x t,
    env_equiv env1 env2 →
    env_equiv (extend env1 x t) (extend env2 x t) :=
begin
  unfold env_equiv extend env.lookup; intros,
  break_if,
end

/- If we extend twice with the same variable, it's equivalent to just
   extending with the second value (i.e. we "overwrite" values). -/
lemma env_equiv_overwrite:
  forall env x t1 t2,
    env_equiv (extend (extend env x t1) x t2)
              (extend env x t2) :=
begin
  unfold env_equiv extend env.lookup; intros,
  break_if,
end

/- If we extend twice with different variables,
   both possible orders result in equivalent envs. -/
lemma env_equiv_neq:
  forall env1 env2 x1 t1 x2 t2,
    x1 ≠ x2 →
    env_equiv env1 env2 →
    env_equiv (extend (extend env1 x1 t1) x2 t2)
              (extend (extend env2 x2 t2) x1 t1) :=
begin
  unfold env_equiv extend env.lookup; intros,
  break_if,
end

lemma env_equiv_lookup_equiv :
  forall env1 env2 x,
    env_equiv env1 env2 →
    env1.lookup x = env2.lookup x :=
begin
  intros,
  unfold env_equiv env.lookup at *,
  rw a,
end

/--

[PROBLEM 5 (5 points)]:

  Prove that if a term is typed in an environment,
  it's typed in equivalent environments.

-/
lemma env_equiv_typed:
  forall env1 ht e t,
    typed env1 ht e t →
    forall env2,
      env_equiv env1 env2 →
      typed env2 ht e t :=
/- About 32 tactics -/
begin
  admit
end
/- END PROBLEM 5 -/

/-
Weakening is a structural property of environments: it means that if a
term has some type in an environment, then the term still has that
type if we extend the environment with a new value for any variable
that is is_free in the term. It's called "weakening" because we can prove
that a term has a type in one environment by removing bindings from
another environment where we know the term is typable.
-/

/-

[PROBLEM 6 (10 points)]:

  Prove weakening for environments.

-/
lemma weaken:
  forall env ht e t,
    typed env ht e t →
    forall x t',
      ¬ is_free e x →
      typed (extend env x t') ht e t :=
/- About 70 tactics -/
begin
  admit
end
/- END PROBLEM 6 -/

/- Next, we'll define another notion of equivalence: equivalence of
environments with respect to the is_free variables in a particular
term. If two environments have the same value for every is_free variable
in a term, then those environments should be interchangeable when
typing that term. -/

def free_env_equiv (E: term) (e1 e2: env) : Prop :=
  forall x,
    is_free E x →
    e1.lookup x = e2.lookup x.

/- We'll prove all the same thigs about this equivalence relation -/
lemma is_free_env_equiv_refl:
  forall E env,
    free_env_equiv E env env :=
begin
  unfold free_env_equiv; intros, trivial,
end

lemma is_free_env_equiv_sym:
  forall E e1 e2,
    free_env_equiv E e1 e2 →
    free_env_equiv E e2 e1 :=
begin
  unfold free_env_equiv; intros,
  simp * at *,
end

lemma is_free_env_equiv_trans:
  forall E e1 e2 e3,
    free_env_equiv E e1 e2 →
    free_env_equiv E e2 e3 →
    free_env_equiv E e1 e3 :=
begin
  unfold free_env_equiv; intros,
  simp * at *,
end

lemma is_free_env_equiv_extend:
  forall E env1 env2 x t,
    free_env_equiv E env1 env2 →
    free_env_equiv E (extend env1 x t) (extend env2 x t) :=
begin
  unfold free_env_equiv extend env.lookup; intros,
  break_if,
end

lemma is_free_env_equiv_overwrite:
  forall E env x t1 t2,
    free_env_equiv E (extend (extend env x t1) x t2)
                     (extend env x t2) :=
begin
  unfold free_env_equiv extend env.lookup; intros,
  break_if,
end

lemma is_free_env_equiv_neq:
  forall E env1 env2 x1 t1 x2 t2,
    x1 ≠ x2 →
    free_env_equiv E env1 env2 →
    free_env_equiv E (extend (extend env1 x1 t1) x2 t2)
                     (extend (extend env2 x2 t2) x1 t1) :=
begin
  unfold free_env_equiv extend env.lookup; intros,
  break_if,
end

/- Here's where we prove interchangeability -/
lemma is_free_env_equiv_typed:
  forall env1 ht e t,
    typed env1 ht e t →
    forall env2,
      free_env_equiv e env1 env2 →
      typed env2 ht e t :=
begin
  unfold free_env_equiv,
  intros env1 ht e t is_typed,
  induction is_typed; intros,
  { constructor },
  { constructor },
  { constructor,
    rw <- a_1, assumption,
    constructor,
  },
  {
    constructor,
    apply ih_1,
    intros,
    apply a_2,
    apply is_free.app_l; assumption,
    apply ih_2,
    intros,
    apply a_2,
    apply is_free.app_r; assumption,
  },
  { constructor,
    apply ih_1; intros,
    unfold stlc_ref.env.lookup stlc_ref.extend,
    break_if,
    apply a_1,
    constructor,
    assumption,
    apply ne.symm,
    delta ne,
    assumption,
  },
  { constructor; assumption, },
  {  constructor,
     apply ih_1; intros,
     apply a_1,
     constructor; assumption,
  },
  { constructor,
     apply ih_1; intros,
     apply a_1,
     constructor; assumption,
  },
  { constructor,
    apply ih_1,
    intros,
    apply a_2,
    apply is_free.assign_l; assumption,
    apply ih_2,
    intros,
    apply a_2,
    apply is_free.assign_r; assumption,
  },
end

/- OK, here's the lemma we needed all that for. The hard part of
proving preservation for the STLC is proving that is_substitution does
the right thing: if it is_substitutes something in for a variable in a
body, the body had better have the same type as if we just assumed
that x had that type. -/

/-

[PROBLEM 7 (10 points)]:

  Prove that is_substitution preserves typing.

-/
lemma is_subst_pres_typed:
  forall e1 e2 x e3,
    is_subst e1 e2 x e3 →
    is_closed e2 →
    forall env ht tA tB,
      typed (extend env x tA) ht e1 tB →
      typed env ht e2 tA →
      typed env ht e3 tB :=
/- About 60 tactics -/
begin
  admit
end
/- END PROBLEM 7 -/

/-- We're almost there. The last thing we'll need to do is to provide
a way to extend heap types with new values during our preservation
proof. To see why, let's take a look at a version of preservation that
looks a lot like the one for STLC without references: -/
lemma wrong_preservation:
  forall h h' e e',
    (h ; e ==> h'; e') →
    is_closed e →
    forall ht t,
      heap_typed ht h →
      typed env.empty ht e t →
      typed env.empty ht e' t := sorry


/- The problem is that we're stuck with the same heap type. This is an
issue because when we allocate a new reference, we'll need to extend
our heap type with a type for the new reference cell. So let's
establish what it means for one heap to extend another: -/

inductive heap_types_extends: heap_types → heap_types → Prop
| nil :
    forall h, heap_types_extends h []
| cons :
    forall h h' x,
      heap_types_extends h' h →
      heap_types_extends (x :: h') (x :: h).

/- We'll need some facts about this definition of heap extension. -/

/- If an address was in our heap type and we extend the heap, we don't
   affect the type for that address. -/

lemma length_cons_lt :
  forall a n (x : a) (h : list a),
  nat.succ n < list.length (x :: h) →
  n < list.length h :=
begin
  intros,
  simp [list.length] at a_1,
  rw add_comm at a_1,
  simp [has_add.add, nat.add] at a_1,
  apply nat.lt_of_succ_lt_succ; assumption,
end
/-

[PROBLEM 8 (5 points)]:

  Prove extends_lookup.

  Hint: `length_cons_lt` will be useful in proving the
  the precondition of the inductive hypothesis.

-/
lemma extends_lookup :
  forall (h h' : heap_types) a,
    heap_types_extends h' h →
    a < list.length h →
    h.lookup a = h'.lookup a :=
/- About 12 tactics -/
begin
  admit
end
/- END PROBLEM 8 -/

/- Extending a heap increases its length -/
lemma length_extends :
  forall h h',
    heap_types_extends h' h →
    forall a,
      a < list.length h →
      a < list.length h' :=
begin
  intros h h' extnds,
  induction extnds; intros,
  { simp [list.length] at a_1,
    cases a_1,
  },
  { simp [list.length] at *,
    rw add_comm at *,
    simp [has_add.add, nat.add] at *,
    cases a_1,
    apply nat.zero_lt_succ,
    apply nat.succ_lt_succ,
    apply ih_1,
    apply nat.lt_of_succ_lt_succ,
    assumption,
  }
end

/- If we snoc a value onto a heap, that extends the heap. -/
lemma extends_snoc :
  forall h x,
    heap_types_extends (snoc h x) h :=
begin
  intros,
  induction h,
  { simp [snoc], constructor },
  { simp [snoc],
    constructor; assumption }
end

/- Heaps extend themselves. -/
lemma extends_refl :
  forall h,
    heap_types_extends h h :=
begin
  induction_on h; constructor; assumption,
end

/-- We'll need anonther version of weakening:
   types are preserved when extending heaps.
-/
lemma heap_weakening :
  forall env ht ht' e t,
    typed env ht e t →
    heap_types_extends ht' ht →
    typed env ht' e t :=
begin
  intros env ht ht' e t e_typed,
  induction e_typed; intros; try { constructor; auto },
  rw extends_lookup,
  constructor,
  apply length_extends; assumption,
  assumption,
  assumption
end

/- It requires a little bit of work to show that extending the heap
   at the end with a new reference preserves the typing of the heap.

   We provide a helper lemma for use below.
-/
lemma heap_typed_snoc :
  forall ht h t e,
    typed env.empty ht e t →
    heap_typed ht h →
    heap_typed (snoc ht t) (snoc h e) :=
begin
  intros,
  unfold heap_typed at *,
  cases a_1, split,
  { simp [length_snoc], cc },
  { intros addr; intros,
    by_cases (addr = list.length ht),
    { subst h,
      apply heap_weakening,
      rw <- left,
      rw heap_lookup_snoc,
      rw left,
      rw heap_types_lookup_snoc,
      assumption,
      apply extends_snoc,
    },
    { rw length_snoc at a_1,
      rw add_comm at a_1,
      simp [has_add.add, nat.add] at a_1,
      have ie := nat.le_of_lt_succ a_1,
      rw le_iff_lt_or_eq at ie,
      cases ie; try { contradiction },
      rw heap_types_lt_lookup_snoc; simp * at *,
      rw heap_lt_lookup_snoc,
      apply heap_weakening,
      apply right,
      simp * at *,
      apply extends_snoc,
      simp * at *,
      simp [left] at *,
      contradiction,
    }
  }
end

/-

OK, we're ready to prove preservation. Note our new theorem
statement: Instead of saying that the term we step to is well-typed
with our starting heap type, we're now saying that there is *some*
heap type which is an extension of our heap type under which the new
term is well-typed. In practice, this new heap type will always either
be the same or the same plus a single type (when a new ref cell is
allocated). You'll need to be careful about when to provide a witness
for this existential, and "eexists" is your friend.

-/


/-

[PROBLEM 9 (30 points)]:

  Prove preservation.

-/
lemma preservation:
  forall h h' e e'
    (e_step : h; e ==> h'; e'),
    is_closed e →
    forall ht t,
      heap_typed ht h →
      typed env.empty ht e t →
      exists ht',
        heap_types_extends ht' ht /\
        typed env.empty ht' e' t /\
        heap_typed ht' h' :=
/- About 153 tactics -/
begin
  admit
end
/- END PROBLEM 9 -/

/- Having proved progress and preservation,
   we're finally ready to prove soundness. -/

/- The empty heap is well-typed in the empty heap type. -/
lemma empty_heap_typed :
  heap_typed heap_types.empty heap.empty :=
begin
  unfold heap_typed; split; intros;
  delta heap.empty heap_types.empty,
  trivial,
  simp [heap.lookup, heap_types.lookup],
  constructor,
end

/- If a term is well-typed in an environment,
   then that environment better have definitions
   for all of that term's is_free variables. -/
lemma typed_is_free_env:
  forall env ht e t,
    typed env ht e t →
    forall x,
      is_free e x →
      exists tx, env.lookup x = some tx :=
begin
  intros env ht e t e_typed,
  induction e_typed; intros,
  { cases a, },
  { cases a, },
  { cases a_1,
    constructor,
    assumption },
  { cases a_2,
    apply ih_1; assumption,
    apply ih_2; assumption,
  },
  { cases a_1,
    specialize (ih_1 _ a_2),
    cases ih_1,
    constructor,
    rw lookup_extend_neq at a_5,
    assumption,
    apply ne.symm; assumption,
  },
  { cases a_2, },
  { cases a_1,
    apply ih_1; assumption
  },
  { cases a_1,
    apply ih_1; assumption, },
  { cases a_2,
    apply ih_1; assumption,
    apply ih_2; assumption, },
end

/-- Therefore, typing in empty env
    implies term is closed. -/
lemma typed_empty_is_closed:
  forall ht e t,
    typed env.empty ht e t →
    is_closed e :=
begin
  unfold is_closed; delta not; intros,
  have X := typed_is_free_env _ _ _ _ a _ a_1,
  cases X, simp [env.empty] at a_3,
  contradiction
end

/- OK, soundness time! -/

/- [PROBLEM 10 (15 points)]: Complete the proof of soundness. -/

/- First, prove that no term which is well-typed in an arbitrary heap
   can get stuck. This proof should use progress and preservation. -/

lemma soundness' :
  forall h e t h' e',
    (h; e ==>* h'; e') →
    forall ht,
      typed env.empty ht e t →
      heap_typed ht h →
      (exists e'' h'', h'; e' ==> h''; e'') \/ is_value e' :=
/- About 12 tactics -/
begin
  admit
end

/- Now, prove that terms which are well-typed
   in the empty heap don't get stuck. -/
lemma soundness :
  forall e t h' e',
    typed env.empty heap_types.empty e t →
    (heap.empty ; e ==>* h'; e') →
    (exists e'' h'', h'; e' ==> h''; e'') \/ is_value e' :=
/- About 4 tactics -/
begin
  admit
end
/- END PROBLEM 10 -/

end stlc_ref
